/*-------------------------------------------------------------*/
/*  File: xbutton.c
 *  V 1.06
 *  Implement a "button window"--a window that displays
 *  a message and calls an "action" routine when the
 *  user presses any mouse button inside the window.
 *  (Uses Xlib's context management routines)
 *
 *  Status: OFF  ---- The action is not called. The message is writed light.
 *          NORMAL 
 *          CROSS_HILITE  -- Hilites when it is crossed.
 *          HILITE -- Hilites.
 *
 *  Exported Routines (are defined in xwins.h):
 *  MakeXButton, HiliteXButton, SetXButtonStatus,
 *  DestroyXButton.
 *
 */
/*-------------------------------------------------------------*/
#include <xwins.h>
#include <xwins_glob.h>

typedef struct D_BUTTON
{
    char     *text;
    void     (*action)(/* caddr_t */);
    caddr_t  action_args;
    int      status;
} D_BUTTON;

void  button_handler(/* XWIN * */);
/*-------------------------------------------------------------*/
/*  M a k e X B u t t o n
 *
 *  A labeled "button" that you can press to perform an "action"
 *
 */
XWIN *MakeXButton(x, y, width, height, bdwidth, bdcolor,
                  bgcolor, parent, text, button_action,
                  action_data)
int           x, y;
unsigned      width, height, bdwidth;
unsigned long bdcolor, bgcolor;
Window        parent; 
char          *text;
void          (*button_action)();
caddr_t       action_data;
{
    XWIN     *new_button;
    D_BUTTON *p_data;
    
/* Allocate button-specific data */
    if((p_data = (D_BUTTON *)calloc(1, sizeof(D_BUTTON))) 
        == NULL)
    {
        fprintf(stderr, "No memory for button's data");
        exit(1);
    } 

/* Allocate a XWIN structure */
    if((new_button = (XWIN *)calloc(1, sizeof(XWIN))) == NULL)
    {
        fprintf(stderr, "No memory for button");
        exit(1);
    } 

/* Initialize button's data and save pointer in new_button */
    p_data->action = button_action;
    p_data->action_args = action_data;
    p_data->text = text;
    p_data->status = NORMAL;
    new_button->data = p_data;
    new_button->event_handler = button_handler;
    new_button->parent = parent;
    new_button->xid = XCreateSimpleWindow(theDisplay, parent,
                      x, y, width, height, bdwidth, bdcolor, 
                      bgcolor);

/* Save new_button as data associated with this window id
 * and the context "xwin_context"
 */
    if(XSaveContext(theDisplay, new_button->xid, xwin_context,
                 (caddr_t) new_button) != 0)
    {
        fprintf(stderr, "Error saving xwin_context data");
        exit(1);
    }
                                
    XSelectInput(theDisplay, new_button->xid, 
        OwnerGrabButtonMask | ExposureMask | ButtonPressMask );
                               
    XMapWindow(theDisplay, new_button->xid);
    return (new_button);
}
/*-------------------------------------------------------------*/
/*  b u t t o n _ h a n d l e r
 *  
 *  Event handler for the "button" 
 */
void button_handler(p_xwin)
XWIN *p_xwin;
{
    D_BUTTON *p_data = (D_BUTTON *) p_xwin->data;
    
/* Handle events occurring in this window */
    if(theEvent.xany.window == p_xwin->xid)
    {
        switch(theEvent.type)
        {
            case Expose:
                if (theEvent.xexpose.count == 0) 
                {
                    XClearWindow(theDisplay, p_xwin->xid);
                    XDrawString(theDisplay, p_xwin->xid,
                        theGC, 
                        theFontStruct->max_bounds.width / 2,
                        theFontStruct->max_bounds.ascent + 
                        theFontStruct->max_bounds.descent,
                        p_data->text, strlen(p_data->text));
		    if(p_data->status == OFF)
		    {
				XFillRectangle(theDisplay, p_xwin->xid,
					LightGC, 0, 0, 1024, 1024);
				XFlush(theDisplay);				
		    }
		    if(p_data->status == HILITE)
				HiliteXButton(p_xwin);
		}
                break;
            case ButtonPress:
/* Call the action routine, PROVIDED it is valid and button is not OFF*/
                if(p_data->action != NULL && p_data->status != OFF)
                        (*p_data->action)(p_data->action_args);
            case EnterNotify:
		if(p_data->status == CROSS_HILITE)
				HiliteXButton(p_xwin);
		break;
	    case LeaveNotify:
		if(p_data->status == CROSS_HILITE)
		{
                    XClearWindow(theDisplay, p_xwin->xid);
                    XDrawString(theDisplay, p_xwin->xid,
                          theGC, 
                        theFontStruct->max_bounds.width / 2,
                        theFontStruct->max_bounds.ascent + 
                        theFontStruct->max_bounds.descent,
                        p_data->text, strlen(p_data->text));
		}
                break;
        }
    }
}
/*-------------------------------------------------------------*/
/*  H i l i t e X B u t t o n
 *
 *  Highlights an XButton
 *
 */
void HiliteXButton(p_button)
XWIN *p_button;
{
	XFillRectangle(theDisplay, p_button->xid, InvertGC,
		0, 0, 1024, 1024);
}
/*-------------------------------------------------------------*/
/*  D e s t r o y X B u t t o n
 *
 *  To properly close a button made by MakeXButton
 */
void DestroyXButton(p_xwin)
XWIN *p_xwin;		/* Pointer to the XWIN structure of the window */
{
	XDestroyWindow(theDisplay, p_xwin->xid);
	XDeleteContext(theDisplay, p_xwin->xid, xwin_context);
	free(p_xwin->data);
	free(p_xwin);
}
/*-----------------------------------------------------------------------*/
/*   S e t X B u t t o n S t a t u s
 *
 *  Changes XButton Status.
 */
void SetXButtonStatus(p_button, status)
XWIN *p_button;
int  status;
{
	D_BUTTON  *p_data;

		p_data = p_button->data;
		p_data->status = status;
		if(status == CROSS_HILITE)
			AddNewEvent(p_button->xid,
					EnterWindowMask|LeaveWindowMask);
}
