/*  File: xwinutil.c
 *  V 2.00-
 *  Utility functions for Xlib programming
 *
 *  Routines Exported (Included in xwins.h):
 * 	xwin_init, GoEventLoop, MakeXWindow, DestroyXWindow, AddNewEvent,
 *	IgnoreEvent, SizeOfString, SizeOfText, DrawText, swaplong,
 *	swapshort, swapbits, system_byte_order, read_cms, write_cms,
 *	two_to, bitsif, SetStandardXColor, SetPixelXColor, GetXColor,
 *      ClearXPixmap.
 */

#define XWIN_UTIL

#include <xwins.h>
#include <xwins_glob.h>

#define LINE_SPACING 2

/* Exported to xtimer */
Atom ATOM_TIMER_EVENT;
void (*time_event_handler)();
caddr_t time_event_data;
/*-----------------------------------------------------------------------*/
/*    x w i n _ i n i t
 *
 * Create a unique context to reference data associated 
 * with the windows
 *
 * Should be called once in order to use the XWIN structures.
 */
void xwin_init()
{
        xwin_context = XUniqueContext();

}
/*-----------------------------------------------------------------------*/
/*   G o E v e n t L o o p
 *
 *   Standard event loop for programs using XWIN structures.
 *   Returns when AppDone=1.
 *   To destroy the main window one has to call XCloseDisplay(theDisplay)
 *   after GoEventLoop returns.
 */
void GoEventLoop(Main_handler,data)
void  (* Main_handler)();            /* Event handler for Main window  */
void   *data;                        /* Pointer to any data needed by  */
                                     /* the handler. Calling sequence: */
                                     /*    Main_handler(data)     */
{
     XWIN  *which_xwin;

/* Event LOOP */
    AppDone=0;
    while (!AppDone) 
    {
/* Waits for coming events */
        XNextEvent(theDisplay, &theEvent);

/* Events for Main window */
        if(theEvent.xany.window == theMain) 
        {
          switch (theEvent.type)
          {
             case FocusIn:           /* Main gets focus */
               if(theEvent.xfocus.mode   == NotifyNormal &&
                  theEvent.xfocus.detail == NotifyNonlinear &&
                  FocusedWindow != theMain )
/* Redirect Focus to subwindow asking for it */
                    XSetInputFocus(theDisplay, FocusedWindow,
                              RevertToNone, CurrentTime);
               break;

             case ClientMessage:
/* Quit signal from WM */
               if(theEvent.xclient.message_type == ATOM_WM_PROTOCOLS &&
                  theEvent.xclient.data.l[0] == ATOM_WM_DELETE_WINDOW)
                         AppDone = 1;
               break;

             case MappingNotify:
/* Keyboard map was changed */
               XRefreshKeyboardMapping((XMappingEvent*)&theEvent);
               break;

             case PropertyNotify:
/* time event */
               if(theEvent.xproperty.atom == ATOM_TIMER_EVENT)
                    time_event_handler(time_event_data);
               break;

             default:
              if(Main_handler != NULL) Main_handler(data);
          }
        }

/* Events for other windows */

/* Retrive the XWIN data */
        if(XFindContext(theDisplay, theEvent.xany.window,
                xwin_context, (caddr_t *) &which_xwin) == 0)
        {
/* If lock is closed, only the window with the key replies */
            if((AppLock == 0)||
                ((AppLock == 1)&& ((which_xwin->xid == AppLockKey)||
                        (which_xwin->parent == AppLockKey))))
            {
/* Call the event handler of this XWIN structure */
                if(*(which_xwin->event_handler) != NULL)
                        (*(which_xwin->event_handler))(which_xwin);
            }

        }
    } /* while */
 
    return;
}


/*-----------------------------------------------------------------------*/
/* M a k e X W i n d o w
 *
 *  Makes a simple window to be drawn. Gives the pointer to the
 *  corresponding XWIN structure.
 *
 */
XWIN *MakeXWindow(x, y, width, height, bdwidth, bdcolor,
                  bgcolor, parent, event_mask, cursor, handler, data)

int           x, y;		/* Position of window (pixels) */
unsigned      width, height;	/* Dimensions of window (pixels) */
unsigned      bdwidth;		/* Width of window border (pixels) */
unsigned long bdcolor, bgcolor;	/* Colors of border an background */
Window        parent;		/* Parent Window */
long          event_mask;	/* Event Mask */
Cursor        cursor;		/* New Cursor if it is not NULL */
void          (*handler)();	/* Handler of events of the window */
void          *data;		/* Pointer to the handler's argument */
{
        XWIN *new_window;

/* Allocate a XWIN structure */
    if((new_window = (XWIN *)calloc(1, sizeof(XWIN))) == NULL)
    {
        fprintf(stderr, "No memory for window\n");
        exit(1);
    }

/* Fill up WIN structure */

        new_window->data = data;
        new_window->event_handler = handler;
        new_window->parent = parent;
        new_window->xid  = XCreateSimpleWindow(theDisplay, parent,
                        x, y, width, height, bdwidth, bdcolor, bgcolor);

/* Save new_window as data associated with this window id
 * and the context "xwin_context"
 */
    if(XSaveContext(theDisplay, new_window->xid, xwin_context,
                 (caddr_t) new_window) != 0)
    {
        fprintf(stderr, "Error saving xwin_context data\n");
        exit(1);
    }
 

/* Set cursor */
        if(cursor != None)
                XDefineCursor(theDisplay, new_window->xid, cursor);

/* Select Input */
        XSelectInput(theDisplay, new_window->xid, event_mask);

/* Map it */
        XMapWindow(theDisplay, new_window->xid);

        return(new_window);
}
/*-----------------------------------------------------------------------*/
/*  D e s t r o y X W i n d o w
 *
 *  To properly close a window made bye MakeXWindow
 */
void DestroyXWindow(p_xwin)
XWIN *p_xwin;		/* Pointer to the XWIN structure of the window */
{
	XUndefineCursor(theDisplay, p_xwin->xid);
	XDestroyWindow(theDisplay, p_xwin->xid);
	XDeleteContext(theDisplay, p_xwin->xid, xwin_context);
	free(p_xwin);
}
/*-----------------------------------------------------------------------*/
/*  A d d N e w E v e n t
 *
 *  Add a new event to the a window's event mask.
 *
 */
void AddNewEvent(xid, event_mask)
Window  xid;		/* XID of the window */
long    event_mask;	/* Event mask to be added */
{
        XWindowAttributes       xwa;
        XSetWindowAttributes    xswa;

        if(XGetWindowAttributes(theDisplay, xid, &xwa) != 0)
        {
                xswa.event_mask = xwa.your_event_mask | event_mask;
                XChangeWindowAttributes(theDisplay, xid, CWEventMask, &xswa);
        }
}
/*-----------------------------------------------------------------------*/
/*  I g n o r e E v e n t
 *
 *  Stop handling an event (by altering the window's mask).
 *
 */
void IgnoreEvent(xid, event_mask)
Window  xid;		/* XID of window */
long    event_mask;	/* Event mask to be substracted */
{
        XWindowAttributes       xwa;
        XSetWindowAttributes    xswa;

        if(XGetWindowAttributes(theDisplay, xid, &xwa) != 0)
        {
                xswa.event_mask = xwa.your_event_mask & (~event_mask);
                XChangeWindowAttributes(theDisplay, xid, CWEventMask, &xswa);
        }
}
/*-----------------------------------------------------------------------*/
/* S i z e O f S t r i n g
 *
 * Determines the width and height (in pixels) of a string
 * using theFontStruct.
 */
void SizeOfString(str, w, h, FS)
XFontStruct *FS;
char *str;		/* String */
unsigned *w,*h;		/* Width and hieght in pixels */
{
     *w = FS->max_bounds.width*strlen(str);
     *h = FS->max_bounds.ascent + FS->max_bounds.descent;
}
/*----------------------------------------------------------------------*/
/* S i z e O f T e x t
 *
 * Determines the width and height (in pixels) of a text of many
 * lines. Each line is an element of an array of strigs. The text is
 * terminated by a null string;
 */

void SizeOfText(text, w, h, FS)
char *text[];		/* Text (array of strings ) */	
unsigned *w,*h;		/* Width and height */
XFontStruct *FS;
{
     int i, ll, llm;

     i   = 0;
     ll  = strlen(text[0]);
     llm = ll;
     while(ll !=0 )
     {
          i++;
          ll = strlen(text[i]);
          if(ll>llm) llm = ll;
     }
     *w = FS->max_bounds.width * llm;
     *h = i*(FS->max_bounds.ascent + 
               FS->max_bounds.descent);
     if(i>0) *h = *h + (i-1)*LINE_SPACING;
}
/*----------------------------------------------------------------------*/
/* D r a w T e x t
 *
 * Draws many lines of text in the position (x,y) of a window.
 * text is an array of strings terminated by a null string.
 */
void DrawText(w, x, y, text)
Window w;		/* XID of window */
int x, y;		/* Position (upper left corner) */
char *text[];		/* Text (array of strings) */
{
     int yy, h, i, ll;

     h = theFontStruct->max_bounds.ascent + 
               theFontStruct->max_bounds.descent + LINE_SPACING;
     i = 0;
     ll = strlen(text[0]);
     while ( ll !=0 )
     {
     yy = y + theFontStruct->max_bounds.ascent + i*h;
     XDrawString(theDisplay, w, theGC, x, yy, text[i], ll);
     i++; ll = strlen(text[i]);
     }
}
/*----------------------------------------------------------------------*/
/* s w a p l o n g
 *
 * To swap bytes of a 4-bytes word
 */
unsigned long swaplong(l)
unsigned long *l;
{
        union{
                unsigned long l;
                char c[4];
        }longint;
        char byte;

                longint.l = *l;
                byte = longint.c[0];
                longint.c[0] = longint.c[3];
                longint.c[3] = byte;
                byte = longint.c[1];
                longint.c[1] = longint.c[2];
                longint.c[2] = byte;
                return longint.l;
}
/*----------------------------------------------------------------------*/
/* s w a p s h o r t
 *
 * To swap bytes of a 2-bytes word
 */
unsigned short swapshort(s)
unsigned short *s;
{
        union{
                unsigned short s;
                char c[2];
        }shortint;
        char byte;

                shortint.s = *s;
                byte = shortint.c[0];
                shortint.c[0] = shortint.c[1];
                shortint.c[1] = byte;
                return shortint.s;
}
/*----------------------------------------------------------------------*/
/* s w a p b i t s
 *
 * To invert the order of the bits of a byte
 */
unsigned char swapbits(c)
char *c;
{
        static unsigned char m[]={1,2,4,8,16,32,64,128};
        unsigned char sc;
        int i;

        sc = 0;
        for(i=0; i<8; i++)
                if((*c)&m[i]) sc = sc|m[7-i];
        return sc;
}
/*----------------------------------------------------------------------*/
/* s y s t e m _ b y t e _ o r d e r
 *
 * Gives the byte order of the system memory: MSBFirst or LSBFirst.
 */
int system_byte_order()
{
        union {short i; char c[2];}one;

/* Determine CPU type */
        one.i =1;
        if(one.c[0]==0) return MSBFirst;
        else    return LSBFirst;
}
/*----------------------------------------------------------------------*/
/* r e a d _ c m s
 *
 * Reads a color map segment from a file.
 * returns:  0: OK, 1: open file error, 2: bad format error
 */
int read_cms(file, size, initial, colors)
char *file;
int *size, *initial;
XColor *colors;
{
	FILE* stream;
	int i;
	unsigned char num, ini, byte;

	stream = fopen(file,"rb");
	if(stream == NULL) return 1;
	if( 1 != fread(&num,1,1,stream))
		{fclose(stream); return 2;}
	if( 1 != fread(&ini,1,1,stream))
		{fclose(stream); return 2;}
	for (i=0; i <= (int)num; i++)
	{
		if(1 != fread(&byte,1,1,stream))
			{fclose(stream); return 2;}
			colors[(int)ini + i].red = 256*(long)byte;
		if(1 != fread(&byte,1,1,stream))
			{fclose(stream); return 2;}
			colors[(int)ini + i].green = 256*(long)byte;
		if(1 != fread(&byte,1,1,stream))
			{fclose(stream); return 2;}
			colors[(int)ini + i].blue = 256*(long)byte;
	}
	fclose(stream);
	*size = (int)num + 1;
	*initial = (int)ini;
	return 0;
}
/*----------------------------------------------------------------------*/
/* w r i t e _ c m s
 *
 * Writes color map segment into a file.
 * returns: 0: OK, 1: open file error, 2: bad format error
 */
int write_cms(file,size,initial,colors)
char* file;
int size,initial;
XColor *colors;
{
	FILE* stream;
	unsigned char byte, rgb[3];
	int i, last;

	stream = fopen(file, "wb");
	if(stream == NULL)return 1;
	if(size <= 0 || initial >= 256)return 2;
	byte = size - 1;
	fwrite(&byte,1,1,stream);
	byte = initial;
	fwrite(&byte,1,1,stream);
	last = size + initial - 1;
	for(i=initial; i<=last; i++)
	{
		rgb[0] = colors[i].red/256;
		rgb[1] = colors[i].green/256;
		rgb[2] = colors[i].blue/256;
		fwrite(rgb, 3, 1, stream);
	}
	fclose(stream);
	return 0;
}
/*----------------------------------------------------------------------*/
/* t w o _ t o
 *
 * Gives the powers of 2
 */
unsigned long two_to(n)
int n;
{
	unsigned long dosn;
	int i;
	dosn = 1;
	for(i=1; i<=n;i++)
	{
		 dosn = dosn*2;
	}
	return dosn;
}
/*----------------------------------------------------------------------*/
/* b i t s i f
 *
 * Gives initial and final non-zero bits of a unsigned long.
 */
void bitsif(m,bi,bf)
unsigned long m;
int *bi, *bf;
{
	unsigned long n,n1;
	*bi = 0;
	n = m;
	if(n == 0)
	{
		*bf = -1;
		return;
	}
	n1 = n/2;
	while (n-n1*2 == 0)
	{
		*bi = *bi +1;
		n = n1;
		n1 = n/2;
	}
	*bf = *bi;
	n = n1;
	n1 = n/2;
	while(n-n1*2 ==1)
	{
		*bf = *bf + 1;
		n = n1;
		n1 = n/2;
	}
}
/*----------------------------------------------------------------------*/
/* S e t S t a n d a r d X C o l o r
 *
 * Fills the rgb fields of an array of XColors with a standard colormap.
 * New algorithm that produces an ordered colormap of 256 elements.
 */
void SetStandardXColor(colors, ncolors)
XColor *colors;
int *ncolors;
{
	long  c[3], rgb[3];
	int i, j, k, l, i0, i1, i2, grey;
	grey = 24;

	i = 2;  /* Index of colormap */
	for(i0=0; i0<3; i0++)
	{
		i1 = (i0+1)/3;
		i1 = i0+1 - i1*3;   /* i0+1 mod 3 */
		i2 = (i0+2)/3;
		i2 = i0+2 - i2*3;   /* i0+2 mod 3 */
/* base dye */
		rgb[i0]=0;
		for(k=0; k<10; k++)
		{
			if( k < 5)
			{
				rgb[i1] = 255;
				rgb[i2] = 51*k;
			}
			else
			{
				rgb[i1] = 51*(10-k);
				rgb[i2] = 255;
			}
/* mixed with white */
			for(j=1;j<3;j++)
			{
				for(l=0;l<3;l++) c[l] = 85*(3-j)+rgb[l]*j/3;
				colors[i].red   = 257*c[0];
				colors[i].green = 257*c[1];
				colors[i].blue  = 257*c[2];
				i++;	
			}
/* mixwd with black */
			
			for(j=0;j<4;j++)
			{
				for(l=0;l<3;l++) c[l] = rgb[l]*(5-j)/5;
				colors[i].red   = 257*c[0];
				colors[i].green = 257*c[1];
				colors[i].blue  = 257*c[2];
				i++;	
			}
/* mixed with grey */
			for(j=1;j<3;j++)
			{
				for(l=0;l<3;l++) c[l] = 24*j+rgb[l]*(3-j)/3;
				colors[i].red   = 257*c[0];
				colors[i].green = 257*c[1];
				colors[i].blue  = 257*c[2];
				i++;	
			}

		}
		
	}
/* Black */
	colors[0].blue  = 0;
	colors[0].green = 0;
	colors[0].red   = 0;
/* White */
	colors[1].blue  = 65535;
	colors[1].green = 65535;
	colors[1].red   = 65535;
/* Grises */
	for(i=1;i<=14;i++)
	{
		colors[241+i].blue  = 4369*i;
		colors[241+i].green = 4369*i;
		colors[241+i].red   = 4369*i;
	}
	*ncolors = 256;
}
/*----------------------------------------------------------------------*/
/* S e t P i x e l X C o l o r
 *
 * Sets the pixel field of a XColor.
 * For TrueColor and DirectColor the masks are used to calculate the pixel.
 * In other cases set pixel is set equal to the index.
 */
void SetPixelXColor(colors, size, initial, offset, visual)
XColor *colors;
int size, initial, offset;
Visual *visual;
{
	int i;
	int Bredi, Bredf, Bgreeni,Bgreenf,Bbluei,Bbluef;
	unsigned long lc;
	if(visual->class == TrueColor || visual->class == DirectColor)
	{
		bitsif(visual->red_mask, &Bredi, &Bredf);
		bitsif(visual->green_mask, &Bgreeni, &Bgreenf);
		bitsif(visual->blue_mask, &Bbluei, &Bbluef);
		for(i=initial; i < initial + size; i++)
		{
		   lc = colors[i].red;
		   colors[i].pixel = two_to(Bredi)*(lc/two_to(15+Bredi-Bredf));
		   lc = colors[i].green;
		   lc = two_to(Bgreeni)*(lc/two_to(15+Bgreeni-Bgreenf));
		   colors[i].pixel = lc | colors[i].pixel;
		   lc = colors[i].blue;
		   lc = two_to(Bbluei)*(lc/two_to(15+Bbluei-Bbluef));
		   colors[i].pixel = colors[i].pixel | lc;
		   colors[i].flags = DoRed | DoGreen | DoBlue;
		}
	}
	else
	{
		for(i=initial; i < size+initial; i++)
		{
			colors[i].pixel = i + offset;
			colors[i].flags = DoRed | DoGreen | DoBlue;
		}
	}
}
/*----------------------------------------------------------------------*/
/* G e t X C o l o r
 *
 * Gets the index of the nearest color of an array of XColor.
 * The imput are the values (0-65535) of red green and blue.
 */
int GetXColor(red, green, blue, colors, n_colors)
unsigned red, green, blue;
int  n_colors;
XColor *colors;
{
	int i, m;
	long int  t, tt;

	tt = 200000; m=-1;
	for(i=0; i<n_colors; i++)
	{
		t = abs(red - colors[i].red);
		t =  t + abs(green - colors[i].green);
		t = t + abs(blue - colors[i].blue);
		if(t < tt)
		{
			tt = t;
			m  = i;
		}
	}
	return m;
}
/*----------------------------------------------------------------------*/
/* C l e a r X P i x m a p 
 *
 *  Fills the given pixmap with the gwBGpix.
 */
void ClearXPixmap(pm, w, h)
unsigned w, h;
Pixmap pm;
{
	XSetForeground(theDisplay, gwGC, gwBGpix);
	XFillRectangle(theDisplay, pm, gwGC, 0, 0, w, h);
	XSetForeground(theDisplay, gwGC, gwFGpix);
}
